#!/usr/bin/env node
import {
  EmmettError,
  isPluginConfig
} from "./chunk-O2VMBOV4.js";

// src/cli.ts
import { Command } from "commander";

// src/commandLine/config.ts
import { Command as CliCommand } from "commander";
import { writeFileSync } from "node:fs";
import { exit } from "process";
var sampleConfig = (plugins = ["emmett-expressjs"]) => {
  const pluginsNames = plugins.length > 0 ? `[
${plugins.map((p) => `"${p}"`).join(",\n")}  
]` : "[]";
  return `
export default {
  plugins: ${pluginsNames},
};
`;
};
var generateConfigFile = (configPath, collectionNames) => {
  try {
    writeFileSync(configPath, sampleConfig(collectionNames), "utf8");
    console.log(`Configuration file stored at: ${configPath}`);
  } catch (error) {
    console.error(`Error: Couldn't store config file: ${configPath}!`);
    console.error(error);
    process.exit(1);
  }
};
var configCommand = new CliCommand("config").description(
  "Manage Pongo configuration"
);
configCommand.command("sample").description("Generate or print sample configuration").option(
  "-plg, --plugins <name>",
  "Specify the plugin name",
  (value, previous) => {
    return previous.concat([value]);
  },
  []
).option(
  "-f, --file <path>",
  "Path to configuration file with collection list"
).option("-g, --generate", "Generate sample config file").option("-p, --print", "Print sample config file").action((options) => {
  const plugins = options.plugin.length > 0 ? options.plugin : ["@event-driven-io/emmett-expressjs"];
  if (!("print" in options) && !("generate" in options)) {
    console.error(
      "Error: Please provide either:\n--print param to print sample config or\n--generate to generate sample config file"
    );
    exit(1);
  }
  if ("print" in options) {
    console.log(`${sampleConfig(plugins)}`);
  } else if ("generate" in options) {
    if (!options.file) {
      console.error(
        "Error: You need to provide a config file through a --file"
      );
      exit(1);
    }
    generateConfigFile(options.file, plugins);
  }
});

// src/commandLine/plugins.ts
import "commander";
import path from "path";
var PluginsConfigImportError = {
  missingDefaultExport: `Error: Config should contain default export, e.g.

${sampleConfig()}`,
  missingPluginsPropertyExport: `Error: Config should contain default export with plugins array, e.g.

${sampleConfig()}`,
  wrongPluginStructure: `Error: Plugin config should be either string with plugin name or object with plugin name, e.g. { name: 'emmett-expressjs' }`
};
var importPluginsConfig = async (options) => {
  const configPath = path.join(
    process.cwd(),
    options?.configPath ?? "./dist/emmett.config.js"
  );
  console.log("IMPORTING" + configPath);
  try {
    const imported = await import(configPath);
    if (!imported.default) {
      return new EmmettError(PluginsConfigImportError.missingDefaultExport);
    }
    if (!imported.default.plugins || !Array.isArray(imported.default.plugins)) {
      return new EmmettError(
        PluginsConfigImportError.missingPluginsPropertyExport
      );
    }
    if (!imported.default.plugins.every(isPluginConfig)) {
      return new EmmettError(PluginsConfigImportError.wrongPluginStructure);
    }
    return { plugins: imported.default.plugins };
  } catch (error) {
    if (!options?.configPath) {
      console.warn("Didn`t find config file: " + configPath, error);
      return { plugins: [] };
    }
    return new EmmettError(
      `Error: Couldn't load file:` + error.toString()
    );
  }
};
var loadPlugins = async (options) => {
  try {
    const pluginsConfig = await importPluginsConfig({
      configPath: options?.configPath
    });
    if (pluginsConfig instanceof EmmettError) throw pluginsConfig;
    if (pluginsConfig.plugins.length === 0) {
      console.log(`No extensions specified in config ${options?.configPath}.`);
      return [];
    }
    const pluginsToLoad = filterPluginsByType(
      pluginsConfig.plugins,
      options?.pluginType
    );
    const pluginsPromises = pluginsToLoad.map(async (pluginConfig) => {
      const importPath = getImportPath(pluginConfig, options?.pluginType);
      try {
        const plugin = await import(importPath);
        console.info(`Loaded plugin: ${importPath}`);
        if (!plugin.default) {
          throw new Error(`Plugin: ${importPath} is missing default export`);
        }
        return plugin.default;
      } catch (error) {
        console.error(`Failed to load extension "${importPath}":`, error);
        return void 0;
      }
    });
    return (await Promise.all(pluginsPromises)).filter(
      (plugin) => plugin !== void 0
    );
  } catch (error) {
    console.error(`Failed to load config ${options?.configPath}:`, error);
    return [];
  }
};
var registerCliPlugins = async (program2, plugins) => {
  const result = [];
  for (const plugin of plugins) {
    const pluginName = plugin.name;
    if (!("registerCommands" in plugin)) {
      console.warn(`No registerCommands function found in ${pluginName}`);
    }
    await plugin.registerCommands(program2);
    console.log(`Loaded extension: ${plugin.name}`);
    result.push(plugin);
  }
};
var filterPluginsByType = (plugins, pluginType) => plugins.filter(
  (p) => typeof p === "string" || pluginType && (p.register === void 0 || p.register.some((r) => r.pluginType === pluginType))
);
var getImportPath = (pluginConfig, pluginType) => {
  if (typeof pluginConfig === "string") {
    return pluginType ? `${pluginConfig}/${pluginType}` : pluginConfig;
  }
  const pluginSubpath = pluginConfig.register.find((r) => pluginType && r.pluginType === pluginType)?.path ?? pluginType;
  return pluginSubpath ? `${pluginConfig.name}/${pluginSubpath}` : pluginConfig.name;
};

// src/cli.ts
var program = new Command();
program.name("emmett").description("CLI tool for Emmett").option("--config <path>", "Path to the configuration file");
var initCLI = async () => {
  const configIndex = process.argv.indexOf("--config");
  const configPath = configIndex !== -1 && process.argv.length > configIndex + 1 ? process.argv[configIndex + 1] : void 0;
  try {
    const plugins = await loadPlugins({
      pluginType: "cli",
      configPath
    });
    await registerCliPlugins(program, plugins);
    program.parse(process.argv);
  } catch (err) {
    console.error(`Failed to load config from ${configPath}:`, err);
  }
};
initCLI().catch((err) => {
  console.error(`CLI initialization failed:`);
  console.error(err);
});
var cli_default = program;
export {
  configCommand,
  cli_default as default,
  generateConfigFile,
  importPluginsConfig,
  loadPlugins,
  registerCliPlugins,
  sampleConfig
};
//# sourceMappingURL=cli.js.map