import {
  ConcurrencyError,
  ConcurrencyInMemoryDatabaseError,
  EmmettError,
  IllegalStateError,
  NotFoundError,
  ValidationError,
  ValidationErrors,
  assertNotEmptyString,
  assertPositiveNumber,
  assertUnsignedBigInt,
  formatDateToUtcYYYYMMDD,
  isBigint,
  isErrorConstructor,
  isNumber,
  isPluginConfig,
  isString,
  isValidYYYYMMDD,
  parseDateFromUtcYYYYMMDD
} from "./chunk-O2VMBOV4.js";

// src/typing/command.ts
var command = (...args) => {
  const [type, data, metadata] = args;
  return metadata !== void 0 ? { type, data, metadata, kind: "Command" } : { type, data, kind: "Command" };
};

// src/typing/event.ts
var event = (...args) => {
  const [type, data, metadata] = args;
  return metadata !== void 0 ? { type, data, metadata, kind: "Event" } : { type, data, kind: "Event" };
};

// src/typing/message.ts
var message = (...args) => {
  const [kind, type, data, metadata] = args;
  return metadata !== void 0 ? { type, data, metadata, kind } : { type, data, kind };
};

// src/typing/workflow.ts
var reply = (message2) => {
  return {
    action: "Reply",
    message: message2
  };
};
var send = (message2) => {
  return {
    action: "Send",
    message: message2
  };
};
var publish = (message2) => {
  return {
    action: "Publish",
    message: message2
  };
};
var schedule = (message2, when) => {
  return {
    action: "Schedule",
    message: message2,
    when
  };
};
var complete = () => {
  return {
    action: "Complete"
  };
};
var ignore = (reason) => {
  return {
    action: "Ignore",
    reason
  };
};
var error = (reason) => {
  return {
    action: "Error",
    reason
  };
};
var accept = () => {
  return { action: "Accept" };
};

// src/eventStore/afterCommit/afterEventStoreCommitHandler.ts
async function tryPublishMessagesAfterCommit(messages, options, context) {
  if (options?.onAfterCommit === void 0) return false;
  try {
    await options?.onAfterCommit(messages, context);
    return true;
  } catch (error2) {
    console.error(`Error in on after commit hook`, error2);
    return false;
  }
}

// src/eventStore/afterCommit/forwardToMessageBus.ts
var forwardToMessageBus = (eventPublisher) => async (messages) => {
  for (const message2 of messages) {
    await eventPublisher.publish(message2);
  }
};

// src/eventStore/events/index.ts
var GlobalStreamCaughtUpType = "__emt:GlobalStreamCaughtUp";
var isGlobalStreamCaughtUp = (event2) => event2.type === GlobalStreamCaughtUpType;
var caughtUpEventFrom = (position) => (event2) => event2.type === GlobalStreamCaughtUpType && event2.metadata?.globalPosition >= position;
var globalStreamCaughtUp = (data) => event(GlobalStreamCaughtUpType, data, {
  globalPosition: data.globalPosition
});
var isSubscriptionEvent = (event2) => isGlobalStreamCaughtUp(event2);
var isNotInternalEvent = (event2) => !isGlobalStreamCaughtUp(event2);

// src/eventStore/eventStore.ts
var canCreateEventStoreSession = (eventStore) => "withSession" in eventStore;
var nulloSessionFactory = (eventStore) => ({
  withSession: (callback) => {
    const nulloSession = {
      eventStore,
      close: () => Promise.resolve()
    };
    return callback(nulloSession);
  }
});

// src/eventStore/expectedVersion.ts
var STREAM_EXISTS = "STREAM_EXISTS";
var STREAM_DOES_NOT_EXIST = "STREAM_DOES_NOT_EXIST";
var NO_CONCURRENCY_CHECK = "NO_CONCURRENCY_CHECK";
var matchesExpectedVersion = (current, expected, defaultVersion) => {
  if (expected === NO_CONCURRENCY_CHECK) return true;
  if (expected == STREAM_DOES_NOT_EXIST) return current === defaultVersion;
  if (expected == STREAM_EXISTS) return current !== defaultVersion;
  return current === expected;
};
var assertExpectedVersionMatchesCurrent = (current, expected, defaultVersion) => {
  expected ??= NO_CONCURRENCY_CHECK;
  if (!matchesExpectedVersion(current, expected, defaultVersion))
    throw new ExpectedVersionConflictError(current, expected);
};
var ExpectedVersionConflictError = class _ExpectedVersionConflictError extends ConcurrencyError {
  constructor(current, expected) {
    super(current?.toString(), expected?.toString());
    Object.setPrototypeOf(this, _ExpectedVersionConflictError.prototype);
  }
};
var isExpectedVersionConflictError = (error2) => error2 instanceof ExpectedVersionConflictError;

// src/eventStore/inMemoryEventStore.ts
import { v4 as uuid5 } from "uuid";

// src/database/inMemoryDatabase.ts
import { v7 as uuid } from "uuid";

// src/utils/collections/duplicates.ts
var hasDuplicates = (array, predicate) => {
  const mapped = array.map(predicate);
  const uniqueValues = new Set(mapped);
  return uniqueValues.size < mapped.length;
};
var getDuplicates = (array, predicate) => {
  const map2 = /* @__PURE__ */ new Map();
  for (let i = 0; i < array.length; i++) {
    const item = array[i];
    const key = predicate(item, i, array);
    if (!map2.has(key)) {
      map2.set(key, []);
    }
    map2.get(key).push(item);
  }
  return Array.from(map2.values()).filter((group) => group.length > 1).flat();
};

// src/utils/collections/merge.ts
var merge = (array, item, where, onExisting, onNotFound = () => void 0) => {
  let wasFound = false;
  const result = array.map((p) => {
    if (!where(p)) return p;
    wasFound = true;
    return onExisting(p);
  }).filter((p) => p !== void 0).map((p) => {
    if (!p) throw Error("That should not happen");
    return p;
  });
  if (!wasFound) {
    const result2 = onNotFound();
    if (result2 !== void 0) return [...array, item];
  }
  return result;
};

// src/utils/collections/index.ts
var arrayUtils = {
  merge,
  hasDuplicates,
  getDuplicates
};

// src/utils/deepEquals.ts
var deepEquals = (left, right) => {
  if (isEquatable(left)) {
    return left.equals(right);
  }
  if (Array.isArray(left)) {
    return Array.isArray(right) && left.length === right.length && left.every((val, index) => deepEquals(val, right[index]));
  }
  if (typeof left !== "object" || typeof right !== "object" || left === null || right === null) {
    return left === right;
  }
  if (Array.isArray(right)) return false;
  const keys1 = Object.keys(left);
  const keys2 = Object.keys(right);
  if (keys1.length !== keys2.length || !keys1.every((key) => keys2.includes(key)))
    return false;
  for (const key in left) {
    if (left[key] instanceof Function && right[key] instanceof Function)
      continue;
    const isEqual = deepEquals(left[key], right[key]);
    if (!isEqual) {
      return false;
    }
  }
  return true;
};
var isEquatable = (left) => {
  return left && typeof left === "object" && "equals" in left && typeof left["equals"] === "function";
};

// src/utils/iterators.ts
var sum = (iterator) => {
  let value, done, sum2 = 0;
  do {
    ({ value, done } = iterator.next());
    sum2 += value || 0;
  } while (!done);
  return sum2;
};

// src/taskProcessing/taskProcessor.ts
var TaskProcessor = class {
  constructor(options) {
    this.options = options;
  }
  queue = [];
  isProcessing = false;
  activeTasks = 0;
  activeGroups = /* @__PURE__ */ new Set();
  enqueue(task, options) {
    if (this.queue.length >= this.options.maxQueueSize) {
      return Promise.reject(
        new EmmettError(
          "Too many pending connections. Please try again later."
        )
      );
    }
    return this.schedule(task, options);
  }
  waitForEndOfProcessing() {
    return this.schedule(({ ack }) => Promise.resolve(ack()));
  }
  schedule(task, options) {
    return promiseWithDeadline(
      (resolve, reject) => {
        const taskWithContext = () => {
          return new Promise((resolveTask, failTask) => {
            const taskPromise = task({
              ack: resolveTask
            });
            taskPromise.then(resolve).catch((err) => {
              failTask(err);
              reject(err);
            });
          });
        };
        this.queue.push({ task: taskWithContext, options });
        if (!this.isProcessing) {
          this.ensureProcessing();
        }
      },
      { deadline: this.options.maxTaskIdleTime }
    );
  }
  ensureProcessing() {
    if (this.isProcessing) return;
    this.isProcessing = true;
    this.processQueue();
  }
  processQueue() {
    try {
      while (this.activeTasks < this.options.maxActiveTasks && this.queue.length > 0) {
        const item = this.takeFirstAvailableItem();
        if (item === null) return;
        const groupId = item.options?.taskGroupId;
        if (groupId) {
          this.activeGroups.add(groupId);
        }
        this.activeTasks++;
        void this.executeItem(item);
      }
    } catch (error2) {
      console.error(error2);
      throw error2;
    } finally {
      this.isProcessing = false;
      if (this.hasItemsToProcess() && this.activeTasks < this.options.maxActiveTasks) {
        this.ensureProcessing();
      }
    }
  }
  async executeItem({ task, options }) {
    try {
      await task();
    } finally {
      this.activeTasks--;
      if (options && options.taskGroupId) {
        this.activeGroups.delete(options.taskGroupId);
      }
      this.ensureProcessing();
    }
  }
  takeFirstAvailableItem = () => {
    const taskIndex = this.queue.findIndex(
      (item2) => !item2.options?.taskGroupId || !this.activeGroups.has(item2.options.taskGroupId)
    );
    if (taskIndex === -1) {
      return null;
    }
    const [item] = this.queue.splice(taskIndex, 1);
    return item ?? null;
  };
  hasItemsToProcess = () => this.queue.findIndex(
    (item) => !item.options?.taskGroupId || !this.activeGroups.has(item.options.taskGroupId)
  ) !== -1;
};
var DEFAULT_PROMISE_DEADLINE = 2147483647;
var promiseWithDeadline = (executor, options) => {
  return new Promise((resolve, reject) => {
    let taskStarted = false;
    const maxWaitingTime = options.deadline || DEFAULT_PROMISE_DEADLINE;
    let timeoutId = setTimeout(() => {
      if (!taskStarted) {
        reject(
          new Error("Task was not started within the maximum waiting time")
        );
      }
    }, maxWaitingTime);
    executor((value) => {
      taskStarted = true;
      if (timeoutId) {
        clearTimeout(timeoutId);
      }
      timeoutId = null;
      resolve(value);
    }, reject);
  });
};

// src/utils/locking/index.ts
var InProcessLock = () => {
  const taskProcessor = new TaskProcessor({
    maxActiveTasks: Number.MAX_VALUE,
    maxQueueSize: Number.MAX_VALUE
  });
  const locks = /* @__PURE__ */ new Map();
  return {
    async acquire({ lockId }) {
      await new Promise((resolve, reject) => {
        taskProcessor.enqueue(
          ({ ack }) => {
            locks.set(lockId, ack);
            resolve();
            return Promise.resolve();
          },
          { taskGroupId: lockId }
        ).catch(reject);
      });
    },
    async tryAcquire({ lockId }) {
      if (locks.has(lockId)) {
        return false;
      }
      await this.acquire({ lockId });
      return true;
    },
    release({ lockId }) {
      const ack = locks.get(lockId);
      if (ack === void 0) {
        return Promise.resolve(true);
      }
      locks.delete(lockId);
      ack();
      return Promise.resolve(true);
    },
    async withAcquire(handle, { lockId }) {
      return taskProcessor.enqueue(
        async ({ ack }) => {
          locks.set(lockId, ack);
          try {
            return await handle();
          } finally {
            locks.delete(lockId);
            ack();
          }
        },
        { taskGroupId: lockId }
      );
    }
  };
};

// src/utils/promises.ts
var delay = (ms) => {
  return new Promise((resolve) => setTimeout(resolve, ms));
};
var asyncAwaiter = () => {
  const result = {};
  (result.reset = () => {
    result.wait = new Promise((res, rej) => {
      result.resolve = res;
      result.reject = rej;
    });
  })();
  return result;
};

// src/utils/retry.ts
import retry from "async-retry";

// src/serialization/json/JSONParser.ts
var ParseError = class extends Error {
  constructor(text) {
    super(`Cannot parse! ${text}`);
  }
};
var JSONParser = {
  stringify: (value, options) => {
    return JSON.stringify(
      options?.map ? options.map(value) : value,
      //TODO: Consider adding support to DateTime and adding specific format to mark that's a bigint
      // eslint-disable-next-line @typescript-eslint/no-unsafe-return
      (_, v) => typeof v === "bigint" ? v.toString() : v
    );
  },
  parse: (text, options) => {
    const parsed = JSON.parse(text, options?.reviver);
    if (options?.typeCheck && !options?.typeCheck(parsed))
      throw new ParseError(text);
    return options?.map ? options.map(parsed) : parsed;
  }
};

// src/utils/retry.ts
var NoRetries = { retries: 0 };
var asyncRetry = async (fn, opts) => {
  if (opts === void 0 || opts.retries === 0) return fn();
  return retry(
    async (bail) => {
      try {
        const result = await fn();
        if (opts?.shouldRetryResult && opts.shouldRetryResult(result)) {
          throw new EmmettError(
            `Retrying because of result: ${JSONParser.stringify(result)}`
          );
        }
        return result;
      } catch (error2) {
        if (opts?.shouldRetryError && !opts.shouldRetryError(error2)) {
          bail(error2);
          return void 0;
        }
        throw error2;
      }
    },
    opts ?? { retries: 0 }
  );
};

// src/database/utils.ts
var isGeneralExpectedDocumentVersion = (version) => {
  return version === "DOCUMENT_DOES_NOT_EXIST" || version === "DOCUMENT_EXISTS" || version === "NO_CONCURRENCY_CHECK";
};
var expectedVersionValue = (version) => version === void 0 || isGeneralExpectedDocumentVersion(version) ? null : version;
var operationResult = (result, options) => {
  const operationResult2 = {
    ...result,
    acknowledged: true,
    successful: result.successful,
    assertSuccessful: (errorMessage) => {
      const { successful } = result;
      const { operationName, collectionName } = options;
      if (!successful)
        throw new ConcurrencyInMemoryDatabaseError(
          errorMessage ?? `${operationName} on ${collectionName} failed. Expected document state does not match current one! Result: ${JSONParser.stringify(result)}!`
        );
    }
  };
  if (options.errors?.throwOnOperationFailures)
    operationResult2.assertSuccessful();
  return operationResult2;
};

// src/database/inMemoryDatabase.ts
var getInMemoryDatabase = () => {
  const storage = /* @__PURE__ */ new Map();
  return {
    collection: (collectionName, collectionOptions = {}) => {
      const ensureCollectionCreated = () => {
        if (!storage.has(collectionName)) storage.set(collectionName, []);
      };
      const errors = collectionOptions.errors;
      const collection = {
        collectionName,
        insertOne: async (document) => {
          ensureCollectionCreated();
          const _id = document._id ?? uuid();
          const _version = document._version ?? 1n;
          const existing = await collection.findOne((c) => c._id === _id);
          if (existing) {
            return operationResult(
              {
                successful: false,
                insertedId: null,
                nextExpectedVersion: _version
              },
              { operationName: "insertOne", collectionName, errors }
            );
          }
          const documentsInCollection = storage.get(collectionName);
          const newDocument = { ...document, _id, _version };
          const newCollection = [...documentsInCollection, newDocument];
          storage.set(collectionName, newCollection);
          return operationResult(
            {
              successful: true,
              insertedId: _id,
              nextExpectedVersion: _version
            },
            { operationName: "insertOne", collectionName, errors }
          );
        },
        findOne: (predicate) => {
          ensureCollectionCreated();
          const documentsInCollection = storage.get(collectionName);
          const filteredDocuments = predicate ? documentsInCollection?.filter((doc) => predicate(doc)) : documentsInCollection;
          const firstOne = filteredDocuments?.[0] ?? null;
          return Promise.resolve(firstOne);
        },
        find: (predicate) => {
          ensureCollectionCreated();
          const documentsInCollection = storage.get(collectionName);
          const filteredDocuments = predicate ? documentsInCollection?.filter((doc) => predicate(doc)) : documentsInCollection;
          return Promise.resolve(filteredDocuments);
        },
        deleteOne: (predicate) => {
          ensureCollectionCreated();
          const documentsInCollection = storage.get(collectionName);
          if (predicate) {
            const foundIndex = documentsInCollection.findIndex(
              (doc) => predicate(doc)
            );
            if (foundIndex === -1) {
              return Promise.resolve(
                operationResult(
                  {
                    successful: false,
                    matchedCount: 0,
                    deletedCount: 0
                  },
                  { operationName: "deleteOne", collectionName, errors }
                )
              );
            } else {
              const newCollection2 = documentsInCollection.toSpliced(
                foundIndex,
                1
              );
              storage.set(collectionName, newCollection2);
              return Promise.resolve(
                operationResult(
                  {
                    successful: true,
                    matchedCount: 1,
                    deletedCount: 1
                  },
                  { operationName: "deleteOne", collectionName, errors }
                )
              );
            }
          }
          const newCollection = documentsInCollection.slice(1);
          storage.set(collectionName, newCollection);
          return Promise.resolve(
            operationResult(
              {
                successful: true,
                matchedCount: 1,
                deletedCount: 1
              },
              { operationName: "deleteOne", collectionName, errors }
            )
          );
        },
        replaceOne: (predicate, document, options) => {
          ensureCollectionCreated();
          const documentsInCollection = storage.get(collectionName);
          const firstIndex = documentsInCollection.findIndex(
            (doc) => predicate(doc)
          );
          if (firstIndex === void 0 || firstIndex === -1) {
            return Promise.resolve(
              operationResult(
                {
                  successful: false,
                  matchedCount: 0,
                  modifiedCount: 0,
                  nextExpectedVersion: 0n
                },
                { operationName: "replaceOne", collectionName, errors }
              )
            );
          }
          const existing = documentsInCollection[firstIndex];
          if (typeof options?.expectedVersion === "bigint" && existing._version !== options.expectedVersion) {
            return Promise.resolve(
              operationResult(
                {
                  successful: false,
                  matchedCount: 1,
                  modifiedCount: 0,
                  nextExpectedVersion: existing._version
                },
                { operationName: "replaceOne", collectionName, errors }
              )
            );
          }
          const newVersion = existing._version + 1n;
          const newCollection = documentsInCollection.with(firstIndex, {
            _id: existing._id,
            ...document,
            _version: newVersion
          });
          storage.set(collectionName, newCollection);
          return Promise.resolve(
            operationResult(
              {
                successful: true,
                modifiedCount: 1,
                matchedCount: firstIndex,
                nextExpectedVersion: newVersion
              },
              { operationName: "replaceOne", collectionName, errors }
            )
          );
        },
        handle: async (id, handle, options) => {
          const { expectedVersion: version, ...operationOptions } = options ?? {};
          ensureCollectionCreated();
          const existing = await collection.findOne(({ _id }) => _id === id);
          const expectedVersion = expectedVersionValue(version);
          if (existing == null && version === "DOCUMENT_EXISTS" || existing == null && expectedVersion != null || existing != null && version === "DOCUMENT_DOES_NOT_EXIST" || existing != null && expectedVersion !== null && existing._version !== expectedVersion) {
            return operationResult(
              {
                successful: false,
                document: existing
              },
              { operationName: "handle", collectionName, errors }
            );
          }
          const result = handle(existing !== null ? { ...existing } : null);
          if (deepEquals(existing, result))
            return operationResult(
              {
                successful: true,
                document: existing
              },
              { operationName: "handle", collectionName, errors }
            );
          if (!existing && result) {
            const newDoc = { ...result, _id: id };
            const insertResult = await collection.insertOne({
              ...newDoc,
              _id: id
            });
            return {
              ...insertResult,
              document: {
                ...newDoc,
                _version: insertResult.nextExpectedVersion
              }
            };
          }
          if (existing && !result) {
            const deleteResult = await collection.deleteOne(
              ({ _id }) => id === _id
            );
            return { ...deleteResult, document: null };
          }
          if (existing && result) {
            const replaceResult = await collection.replaceOne(
              ({ _id }) => id === _id,
              result,
              {
                ...operationOptions,
                expectedVersion: expectedVersion ?? "DOCUMENT_EXISTS"
              }
            );
            return {
              ...replaceResult,
              document: {
                ...result,
                _version: replaceResult.nextExpectedVersion
              }
            };
          }
          return operationResult(
            {
              successful: true,
              document: existing
            },
            { operationName: "handle", collectionName, errors }
          );
        }
      };
      return collection;
    }
  };
};

// src/eventStore/projections/inMemory/inMemoryProjection.ts
var DATABASE_REQUIRED_ERROR_MESSAGE = "Database is required in context for InMemory projections";
var handleInMemoryProjections = async (options) => {
  const { projections: projections2, events, database, eventStore } = options;
  const eventTypes = events.map((e) => e.type);
  const relevantProjections = projections2.filter(
    (p) => p.canHandle.some((type) => eventTypes.includes(type))
  );
  for (const projection2 of relevantProjections) {
    await projection2.handle(events, {
      eventStore,
      database
    });
  }
};
var inMemoryProjection = ({
  truncate,
  handle,
  canHandle
}) => ({
  canHandle,
  handle: async (events, context) => {
    if (!context.database) {
      throw new Error(DATABASE_REQUIRED_ERROR_MESSAGE);
    }
    await handle(events, {
      ...context,
      database: context.database
    });
  },
  truncate: truncate ? (context) => {
    if (!context.database) {
      throw new Error(DATABASE_REQUIRED_ERROR_MESSAGE);
    }
    return truncate({
      ...context,
      database: context.database
    });
  } : void 0
});
var inMemoryMultiStreamProjection = (options) => {
  const { collectionName, getDocumentId, canHandle } = options;
  return inMemoryProjection({
    handle: async (events, { database }) => {
      const collection = database.collection(collectionName);
      for (const event2 of events) {
        await collection.handle(getDocumentId(event2), (document) => {
          if ("initialState" in options) {
            return options.evolve(document ?? options.initialState(), event2);
          } else {
            return options.evolve(document, event2);
          }
        });
      }
    },
    canHandle,
    truncate: async ({
      database
    }) => {
      const collection = database.collection(collectionName);
      const documents = await collection.find();
      for (const doc of documents) {
        if (doc && "_id" in doc) {
          const id = doc._id;
          await collection.deleteOne((d) => d._id === id);
        }
      }
    }
  });
};
var inMemorySingleStreamProjection = (options) => {
  return inMemoryMultiStreamProjection({
    ...options,
    getDocumentId: options.getDocumentId ?? ((event2) => event2.metadata.streamName)
  });
};

// src/eventStore/projections/inMemory/inMemoryProjectionSpec.ts
import { v4 as uuid2 } from "uuid";

// src/testing/assertions.ts
var AssertionError = class extends Error {
  constructor(message2) {
    super(message2);
  }
};
var isSubset = (superObj, subObj) => {
  const sup = superObj;
  const sub = subObj;
  assertOk(sup);
  assertOk(sub);
  return Object.keys(sub).every((ele) => {
    if (typeof sub[ele] == "object") {
      return isSubset(sup[ele], sub[ele]);
    }
    return sub[ele] === sup[ele];
  });
};
var assertFails = (message2) => {
  throw new AssertionError(message2 ?? "That should not ever happened, right?");
};
var assertThrowsAsync = async (fun, errorCheck) => {
  try {
    await fun();
  } catch (error2) {
    const typedError = error2;
    if (typedError instanceof AssertionError || !errorCheck) {
      assertFalse(
        typedError instanceof AssertionError,
        "Function didn't throw expected error"
      );
      return typedError;
    }
    assertTrue(
      errorCheck(typedError),
      `Error doesn't match the expected condition: ${JSONParser.stringify(error2)}`
    );
    return typedError;
  }
  throw new AssertionError("Function didn't throw expected error");
};
var assertThrows = (fun, errorCheck) => {
  try {
    fun();
  } catch (error2) {
    const typedError = error2;
    if (errorCheck) {
      assertTrue(
        errorCheck(typedError),
        `Error doesn't match the expected condition: ${JSONParser.stringify(error2)}`
      );
    } else if (typedError instanceof AssertionError) {
      assertFalse(
        typedError instanceof AssertionError,
        "Function didn't throw expected error"
      );
    }
    return typedError;
  }
  throw new AssertionError("Function didn't throw expected error");
};
var assertDoesNotThrow = (fun, errorCheck) => {
  try {
    fun();
    return null;
  } catch (error2) {
    const typedError = error2;
    if (errorCheck) {
      assertFalse(
        errorCheck(typedError),
        `Error matching the expected condition was thrown!: ${JSONParser.stringify(error2)}`
      );
    } else {
      assertFails(`Function threw an error: ${JSONParser.stringify(error2)}`);
    }
    return typedError;
  }
};
var assertRejects = async (promise, errorCheck) => {
  try {
    await promise;
    throw new AssertionError("Function didn't throw expected error");
  } catch (error2) {
    if (!errorCheck) return;
    if (errorCheck instanceof Error) assertDeepEqual(error2, errorCheck);
    else assertTrue(errorCheck(error2));
  }
};
var assertMatches = (actual, expected, message2) => {
  if (!isSubset(actual, expected))
    throw new AssertionError(
      message2 ?? `subObj:
${JSONParser.stringify(expected)}
is not subset of
${JSONParser.stringify(actual)}`
    );
};
var assertDeepEqual = (actual, expected, message2) => {
  if (!deepEquals(actual, expected))
    throw new AssertionError(
      message2 ?? `subObj:
${JSONParser.stringify(expected)}
is not equal to
${JSONParser.stringify(actual)}`
    );
};
var assertNotDeepEqual = (actual, expected, message2) => {
  if (deepEquals(actual, expected))
    throw new AssertionError(
      message2 ?? `subObj:
${JSONParser.stringify(expected)}
is equals to
${JSONParser.stringify(actual)}`
    );
};
var assertThat = (item) => {
  return {
    isEqualTo: (other) => assertTrue(deepEquals(item, other))
  };
};
function assertFalse(condition, message2) {
  if (condition !== false)
    throw new AssertionError(message2 ?? `Condition is true`);
}
function assertTrue(condition, message2) {
  if (condition !== true)
    throw new AssertionError(message2 ?? `Condition is false`);
}
function assertOk(obj, message2) {
  if (!obj) throw new AssertionError(message2 ?? `Condition is not truthy`);
}
function assertEqual(expected, actual, message2) {
  if (expected !== actual)
    throw new AssertionError(
      `${message2 ?? "Objects are not equal"}:
Expected: ${JSONParser.stringify(expected)}
Actual: ${JSONParser.stringify(actual)}`
    );
}
function assertNotEqual(obj, other, message2) {
  if (obj === other)
    throw new AssertionError(
      message2 ?? `Objects are equal: ${JSONParser.stringify(obj)}`
    );
}
function assertIsNotNull(result) {
  assertNotEqual(result, null);
  assertOk(result);
}
function assertIsNull(result) {
  assertEqual(result, null);
}
var argValue = (value) => (arg) => deepEquals(arg, value);
var argMatches = (matches) => (arg) => matches(arg);
function verifyThat(fn) {
  return {
    calledTimes: (times) => {
      assertEqual(fn.mock?.calls?.length, times);
    },
    notCalled: () => {
      assertEqual(fn?.mock?.calls?.length, 0);
    },
    called: () => {
      assertTrue(
        fn.mock?.calls.length !== void 0 && fn.mock.calls.length > 0
      );
    },
    calledWith: (...args) => {
      assertTrue(
        fn.mock?.calls.length !== void 0 && fn.mock.calls.length >= 1 && fn.mock.calls.some((call) => deepEquals(call.arguments, args))
      );
    },
    calledOnceWith: (...args) => {
      assertTrue(
        fn.mock?.calls.length !== void 0 && fn.mock.calls.length === 1 && fn.mock.calls.some((call) => deepEquals(call.arguments, args))
      );
    },
    calledWithArgumentMatching: (...matches) => {
      assertTrue(
        fn.mock?.calls.length !== void 0 && fn.mock.calls.length >= 1
      );
      assertTrue(
        fn.mock?.calls.length !== void 0 && fn.mock.calls.length >= 1 && fn.mock.calls.some(
          (call) => call.arguments && call.arguments.length >= matches.length && matches.every((match, index) => match(call.arguments[index]))
        )
      );
    },
    notCalledWithArgumentMatching: (...matches) => {
      assertFalse(
        fn.mock?.calls.length !== void 0 && fn.mock.calls.length >= 1 && fn.mock.calls[0].arguments && fn.mock.calls[0].arguments.length >= matches.length && matches.every(
          (match, index) => match(fn.mock.calls[0].arguments[index])
        )
      );
    }
  };
}
var assertThatArray = (array) => {
  return {
    isEmpty: () => assertEqual(
      array.length,
      0,
      `Array is not empty ${JSONParser.stringify(array)}`
    ),
    isNotEmpty: () => assertNotEqual(array.length, 0, `Array is empty`),
    hasSize: (length) => assertEqual(array.length, length),
    containsElements: (other) => {
      assertTrue(other.every((ts) => array.some((o) => deepEquals(ts, o))));
    },
    containsElementsMatching: (other) => {
      assertTrue(other.every((ts) => array.some((o) => isSubset(o, ts))));
    },
    containsOnlyElementsMatching: (other) => {
      assertEqual(array.length, other.length, `Arrays lengths don't match`);
      assertTrue(other.every((ts) => array.some((o) => isSubset(o, ts))));
    },
    containsExactlyInAnyOrder: (other) => {
      assertEqual(array.length, other.length);
      assertTrue(array.every((ts) => other.some((o) => deepEquals(ts, o))));
    },
    containsExactlyInAnyOrderElementsOf: (other) => {
      assertEqual(array.length, other.length);
      assertTrue(array.every((ts) => other.some((o) => deepEquals(ts, o))));
    },
    containsExactlyElementsOf: (other) => {
      assertEqual(array.length, other.length);
      for (let i = 0; i < array.length; i++) {
        assertTrue(deepEquals(array[i], other[i]));
      }
    },
    containsExactly: (elem) => {
      assertEqual(array.length, 1);
      assertTrue(deepEquals(array[0], elem));
    },
    contains: (elem) => {
      assertTrue(array.some((a) => deepEquals(a, elem)));
    },
    containsOnlyOnceElementsOf: (other) => {
      assertTrue(
        other.map((o) => array.filter((a) => deepEquals(a, o)).length).filter((a) => a === 1).length === other.length
      );
    },
    containsAnyOf: (other) => {
      assertTrue(array.some((a) => other.some((o) => deepEquals(a, o))));
    },
    allMatch: (matches) => {
      assertTrue(array.every(matches));
    },
    anyMatches: (matches) => {
      assertTrue(array.some(matches));
    },
    allMatchAsync: async (matches) => {
      for (const item of array) {
        assertTrue(await matches(item));
      }
    }
  };
};

// src/testing/deciderSpecification.ts
var DeciderSpecification = {
  for: deciderSpecificationFor
};
function deciderSpecificationFor(decider) {
  {
    return (givenEvents) => {
      return {
        when: (command2) => {
          const handle = () => {
            const existingEvents = Array.isArray(givenEvents) ? givenEvents : [givenEvents];
            const currentState = existingEvents.reduce(
              decider.evolve,
              decider.initialState()
            );
            return decider.decide(command2, currentState);
          };
          return {
            then: (expectedEvents) => {
              const resultEvents = handle();
              if (resultEvents instanceof Promise) {
                return resultEvents.then((events) => {
                  thenHandler(events, expectedEvents);
                });
              }
              thenHandler(resultEvents, expectedEvents);
            },
            thenNothingHappened: () => {
              const resultEvents = handle();
              if (resultEvents instanceof Promise) {
                return resultEvents.then((events) => {
                  thenNothingHappensHandler(events);
                });
              }
              thenNothingHappensHandler(resultEvents);
            },
            thenThrows: (...args) => {
              try {
                const result = handle();
                if (result instanceof Promise) {
                  return result.then(() => {
                    throw new AssertionError(
                      "Handler did not fail as expected"
                    );
                  }).catch((error2) => {
                    thenThrowsErrorHandler(error2, args);
                  });
                }
                throw new AssertionError("Handler did not fail as expected");
              } catch (error2) {
                thenThrowsErrorHandler(error2, args);
              }
            }
          };
        }
      };
    };
  }
}
function thenHandler(events, expectedEvents) {
  const resultEventsArray = Array.isArray(events) ? events : [events];
  const expectedEventsArray = Array.isArray(expectedEvents) ? expectedEvents : [expectedEvents];
  assertThatArray(resultEventsArray).containsOnlyElementsMatching(
    expectedEventsArray
  );
}
function thenNothingHappensHandler(events) {
  const resultEventsArray = Array.isArray(events) ? events : [events];
  assertThatArray(resultEventsArray).isEmpty();
}
function thenThrowsErrorHandler(error2, args) {
  if (error2 instanceof AssertionError) throw error2;
  if (args.length === 0) return;
  if (!isErrorConstructor(args[0])) {
    assertTrue(
      args[0](error2),
      `Error didn't match the error condition: ${error2?.toString()}`
    );
    return;
  }
  assertTrue(
    error2 instanceof args[0],
    `Caught error is not an instance of the expected type: ${error2?.toString()}`
  );
  if (args[1]) {
    assertTrue(
      args[1](error2),
      `Error didn't match the error condition: ${error2?.toString()}`
    );
  }
}

// src/testing/wrapEventStore.ts
var WrapEventStore = (eventStore) => {
  const appendedEvents = /* @__PURE__ */ new Map();
  const wrapped = {
    ...eventStore,
    aggregateStream(streamName, options) {
      return eventStore.aggregateStream(streamName, options);
    },
    async readStream(streamName, options) {
      return await eventStore.readStream(
        streamName,
        options
      );
    },
    appendToStream: async (streamName, events, options) => {
      const result = await eventStore.appendToStream(
        streamName,
        events,
        options
      );
      const currentStream = appendedEvents.get(streamName) ?? [streamName, []];
      appendedEvents.set(streamName, [
        streamName,
        [...currentStream[1], ...events]
      ]);
      return result;
    },
    appendedEvents,
    setup: async (streamName, events) => {
      return eventStore.appendToStream(streamName, events);
    }
    // streamEvents: (): ReadableStream<
    //   // eslint-disable-next-line @typescript-eslint/no-redundant-type-constituents
    //   ReadEvent<Event, ReadEventMetadataType> | GlobalSubscriptionEvent
    // > => {
    //   return eventStore.streamEvents();
    // },
  };
  return wrapped;
};

// src/eventStore/projections/inMemory/inMemoryProjectionSpec.ts
var InMemoryProjectionSpec = {
  for: (options) => {
    const { projection: projection2 } = options;
    return (givenEvents) => {
      return {
        when: (events, options2) => {
          const allEvents = [];
          const run = async (database) => {
            let globalPosition = 0n;
            const numberOfTimes = options2?.numberOfTimes ?? 1;
            for (const event2 of [
              ...givenEvents,
              ...Array.from({ length: numberOfTimes }).flatMap(() => events)
            ]) {
              const metadata = {
                globalPosition: ++globalPosition,
                streamPosition: globalPosition,
                streamName: event2.metadata?.streamName ?? `test-${uuid2()}`,
                messageId: uuid2()
              };
              allEvents.push({
                ...event2,
                kind: "Event",
                metadata: {
                  ...metadata,
                  ..."metadata" in event2 ? event2.metadata ?? {} : {}
                }
              });
            }
            const mockEventStore = {
              database,
              aggregateStream: async () => {
                return Promise.resolve({
                  state: {},
                  currentStreamVersion: 0n,
                  streamExists: false
                });
              },
              readStream: async () => {
                return Promise.resolve({
                  events: [],
                  currentStreamVersion: 0n,
                  streamExists: false
                });
              },
              appendToStream: async () => {
                return Promise.resolve({
                  nextExpectedStreamVersion: 0n,
                  createdNewStream: false
                });
              }
            };
            await handleInMemoryProjections({
              events: allEvents,
              projections: [projection2],
              database,
              eventStore: mockEventStore
            });
          };
          return {
            then: async (assertFn, message2) => {
              const database = getInMemoryDatabase();
              await run(database);
              const succeeded = await assertFn({ database });
              if (succeeded !== void 0 && succeeded === false) {
                assertFails(
                  message2 ?? "Projection specification didn't match the criteria"
                );
              }
            },
            thenThrows: async (...args) => {
              const database = getInMemoryDatabase();
              try {
                await run(database);
                throw new AssertionError("Handler did not fail as expected");
              } catch (error2) {
                if (error2 instanceof AssertionError) throw error2;
                if (args.length === 0) return;
                if (!isErrorConstructor(args[0])) {
                  assertTrue(
                    args[0](error2),
                    `Error didn't match the error condition: ${error2?.toString()}`
                  );
                  return;
                }
                assertTrue(
                  error2 instanceof args[0],
                  `Caught error is not an instance of the expected type: ${error2?.toString()}`
                );
                if (args[1]) {
                  assertTrue(
                    args[1](error2),
                    `Error didn't match the error condition: ${error2?.toString()}`
                  );
                }
              }
            }
          };
        }
      };
    };
  }
};
var eventInStream = (streamName, event2) => {
  return {
    ...event2,
    metadata: {
      ...event2.metadata ?? {},
      streamName: event2.metadata?.streamName ?? streamName
    }
  };
};
var eventsInStream = (streamName, events) => {
  return events.map((e) => eventInStream(streamName, e));
};
var newEventsInStream = eventsInStream;
function documentExists(expected, options) {
  return async ({ database }) => {
    const collection = database.collection(options.inCollection);
    const document = await collection.findOne((doc) => {
      const docId = "_id" in doc ? doc._id : void 0;
      return docId === options.withId;
    });
    if (!document) {
      assertFails(
        `Document with ID ${options.withId} does not exist in collection ${options.inCollection}`
      );
      return Promise.resolve(false);
    }
    for (const [key, value] of Object.entries(expected)) {
      const propKey = key;
      if (!(key in document) || JSONParser.stringify(document[propKey]) !== JSONParser.stringify(value)) {
        assertFails(`Property ${key} doesn't match the expected value`);
        return Promise.resolve(false);
      }
    }
    return Promise.resolve(true);
  };
}
var expectInMemoryDocuments = {
  fromCollection: (collectionName) => ({
    withId: (id) => ({
      toBeEqual: (expected) => documentExists(expected, {
        inCollection: collectionName,
        withId: id
      })
    })
  })
};

// src/eventStore/subscriptions/caughtUpTransformStream.ts
import { TransformStream } from "web-streams-polyfill";
var streamTrackingGlobalPosition = (currentEvents) => new CaughtUpTransformStream(currentEvents);
var CaughtUpTransformStream = class extends TransformStream {
  _currentPosition;
  _logPosition;
  constructor(events) {
    super({
      start: (controller) => {
        let globalPosition = 0n;
        for (const event2 of events) {
          controller.enqueue(event2);
          globalPosition = event2.metadata.globalPosition;
        }
        controller.enqueue(globalStreamCaughtUp({ globalPosition }));
      },
      transform: (event2, controller) => {
        this._currentPosition = event2.metadata.globalPosition;
        controller.enqueue(event2);
        if (this._currentPosition < this._logPosition) return;
        controller.enqueue(
          globalStreamCaughtUp({ globalPosition: this._currentPosition })
        );
      }
    });
    this._currentPosition = this._logPosition = events.length > 0 ? events[events.length - 1].metadata.globalPosition : 0n;
  }
  set logPosition(value) {
    this._logPosition = value;
  }
};

// src/eventStore/subscriptions/streamingCoordinator.ts
import { v4 as uuid4 } from "uuid";

// src/streaming/transformations/notifyAboutNoActiveReaders.ts
import { v4 as uuid3 } from "uuid";
import { TransformStream as TransformStream2 } from "web-streams-polyfill";
var notifyAboutNoActiveReadersStream = (onNoActiveReaderCallback, options = {}) => new NotifyAboutNoActiveReadersStream(onNoActiveReaderCallback, options);
var NotifyAboutNoActiveReadersStream = class extends TransformStream2 {
  constructor(onNoActiveReaderCallback, options = {}) {
    super({
      cancel: (reason) => {
        console.log("Stream was canceled. Reason:", reason);
        this.stopChecking();
      }
    });
    this.onNoActiveReaderCallback = onNoActiveReaderCallback;
    this.streamId = options?.streamId ?? uuid3();
    this.onNoActiveReaderCallback = onNoActiveReaderCallback;
    this.startChecking(options?.intervalCheckInMs ?? 20);
  }
  checkInterval = null;
  streamId;
  _isStopped = false;
  get hasActiveSubscribers() {
    return !this._isStopped;
  }
  startChecking(interval) {
    this.checkInterval = setInterval(() => {
      this.checkNoActiveReader();
    }, interval);
  }
  stopChecking() {
    if (!this.checkInterval) return;
    clearInterval(this.checkInterval);
    this.checkInterval = null;
    this._isStopped = true;
    this.onNoActiveReaderCallback(this);
  }
  checkNoActiveReader() {
    if (!this.readable.locked && !this._isStopped) {
      this.stopChecking();
    }
  }
};

// src/streaming/writers/writeToStream.ts
var writeToStream = async (stream, items) => {
  if (stream.writable.locked) return false;
  const writer = stream.writable.getWriter();
  await writer.ready;
  if (!stream.readable.locked) return false;
  try {
    for (const item of items) {
      await writer.write(item);
    }
  } catch (error2) {
    console.log(error2);
  } finally {
    await writer.close();
  }
  return true;
};

// src/eventStore/subscriptions/streamingCoordinator.ts
var StreamingCoordinator = () => {
  const allEvents = [];
  const listeners = /* @__PURE__ */ new Map();
  return {
    notify: async (events) => {
      if (events.length === 0) return;
      allEvents.push(...events);
      for (const listener of listeners.values()) {
        listener.logPosition = events[events.length - 1].metadata.globalPosition;
        await writeToStream(listener, events);
      }
    },
    stream: () => {
      const streamId = uuid4();
      const transformStream = streamTrackingGlobalPosition(allEvents);
      listeners.set(streamId, transformStream);
      return transformStream.readable.pipeThrough(
        notifyAboutNoActiveReadersStream(
          (stream) => {
            if (listeners.has(stream.streamId))
              listeners.delete(stream.streamId);
          },
          { streamId }
        )
      );
    }
  };
};

// src/eventStore/inMemoryEventStore.ts
var InMemoryEventStoreDefaultStreamVersion = 0n;
var getInMemoryEventStore = (eventStoreOptions) => {
  const streams = /* @__PURE__ */ new Map();
  const streamingCoordinator = StreamingCoordinator();
  const getAllEventsCount = () => {
    return Array.from(streams.values()).map((s) => s.length).reduce((p, c) => p + c, 0);
  };
  const database = eventStoreOptions?.database || getInMemoryDatabase();
  const inlineProjections2 = (eventStoreOptions?.projections ?? []).filter(({ type }) => type === "inline").map(({ projection: projection2 }) => projection2);
  const eventStore = {
    database,
    async aggregateStream(streamName, options) {
      const { evolve, initialState, read } = options;
      const result = await this.readStream(streamName, read);
      const events = result?.events ?? [];
      return {
        currentStreamVersion: BigInt(events.length),
        state: events.reduce(evolve, initialState()),
        streamExists: result.streamExists
      };
    },
    readStream: (streamName, options) => {
      const events = streams.get(streamName);
      const currentStreamVersion = events ? BigInt(events.length) : InMemoryEventStoreDefaultStreamVersion;
      assertExpectedVersionMatchesCurrent(
        currentStreamVersion,
        options?.expectedStreamVersion,
        InMemoryEventStoreDefaultStreamVersion
      );
      const from = Number(options && "from" in options ? options.from : 0);
      const to = Number(
        options && "to" in options ? options.to : options && "maxCount" in options && options.maxCount ? options.from + options.maxCount : events?.length ?? 1
      );
      const resultEvents = events !== void 0 && events.length > 0 ? events.map(
        (e) => e
      ).slice(from, to) : [];
      const result = {
        currentStreamVersion,
        events: resultEvents,
        streamExists: events !== void 0 && events.length > 0
      };
      return Promise.resolve(result);
    },
    appendToStream: async (streamName, events, options) => {
      const currentEvents = streams.get(streamName) ?? [];
      const currentStreamVersion = currentEvents.length > 0 ? BigInt(currentEvents.length) : InMemoryEventStoreDefaultStreamVersion;
      assertExpectedVersionMatchesCurrent(
        currentStreamVersion,
        options?.expectedStreamVersion,
        InMemoryEventStoreDefaultStreamVersion
      );
      const newEvents = events.map((event2, index) => {
        const metadata = {
          streamName,
          messageId: uuid5(),
          streamPosition: BigInt(currentEvents.length + index + 1),
          globalPosition: BigInt(getAllEventsCount() + index + 1)
        };
        return {
          ...event2,
          kind: event2.kind ?? "Event",
          metadata: {
            ..."metadata" in event2 ? event2.metadata ?? {} : {},
            ...metadata
          }
        };
      });
      const positionOfLastEventInTheStream = BigInt(
        newEvents.slice(-1)[0].metadata.streamPosition
      );
      streams.set(streamName, [...currentEvents, ...newEvents]);
      await streamingCoordinator.notify(newEvents);
      if (inlineProjections2.length > 0) {
        await handleInMemoryProjections({
          projections: inlineProjections2,
          events: newEvents,
          database: eventStore.database,
          eventStore
        });
      }
      const result = {
        nextExpectedStreamVersion: positionOfLastEventInTheStream,
        createdNewStream: currentStreamVersion === InMemoryEventStoreDefaultStreamVersion
      };
      await tryPublishMessagesAfterCommit(
        newEvents,
        eventStoreOptions?.hooks
      );
      return result;
    }
    //streamEvents: streamingCoordinator.stream,
  };
  return eventStore;
};

// src/commandHandling/handleCommand.ts
var CommandHandlerStreamVersionConflictRetryOptions = {
  retries: 3,
  minTimeout: 100,
  factor: 1.5,
  shouldRetryError: isExpectedVersionConflictError
};
var fromCommandHandlerRetryOptions = (retryOptions) => {
  if (retryOptions === void 0) return NoRetries;
  if ("onVersionConflict" in retryOptions) {
    if (typeof retryOptions.onVersionConflict === "boolean")
      return CommandHandlerStreamVersionConflictRetryOptions;
    else if (typeof retryOptions.onVersionConflict === "number")
      return {
        ...CommandHandlerStreamVersionConflictRetryOptions,
        retries: retryOptions.onVersionConflict
      };
    else return retryOptions.onVersionConflict;
  }
  return retryOptions;
};
var CommandHandler = (options) => async (store, id, handle, handleOptions) => asyncRetry(
  async () => {
    const result = await withSession(store, async ({ eventStore }) => {
      const { evolve, initialState } = options;
      const mapToStreamId = options.mapToStreamId ?? ((id2) => id2);
      const streamName = mapToStreamId(id);
      const aggregationResult = await eventStore.aggregateStream(streamName, {
        evolve,
        initialState,
        read: {
          // expected stream version is passed to fail fast
          // if stream is in the wrong state
          expectedStreamVersion: handleOptions?.expectedStreamVersion ?? NO_CONCURRENCY_CHECK
        }
      });
      const {
        // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
        currentStreamVersion,
        streamExists: _streamExists,
        ...restOfAggregationResult
      } = aggregationResult;
      let state = aggregationResult.state;
      const handlers = Array.isArray(handle) ? handle : [handle];
      let eventsToAppend = [];
      for (const handler of handlers) {
        const result2 = await handler(state);
        const newEvents = Array.isArray(result2) ? result2 : [result2];
        if (newEvents.length > 0) {
          state = newEvents.reduce(evolve, state);
        }
        eventsToAppend = [...eventsToAppend, ...newEvents];
      }
      if (eventsToAppend.length === 0) {
        return {
          ...restOfAggregationResult,
          newEvents: [],
          newState: state,
          // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
          nextExpectedStreamVersion: currentStreamVersion,
          createdNewStream: false
        };
      }
      const expectedStreamVersion = handleOptions?.expectedStreamVersion ?? (aggregationResult.streamExists ? currentStreamVersion : STREAM_DOES_NOT_EXIST);
      const appendResult = await eventStore.appendToStream(
        streamName,
        eventsToAppend,
        {
          ...handleOptions,
          expectedStreamVersion
        }
      );
      return {
        ...appendResult,
        newEvents: eventsToAppend,
        newState: state
      };
    });
    return result;
  },
  fromCommandHandlerRetryOptions(
    handleOptions && "retry" in handleOptions ? handleOptions.retry : options.retry
  )
);
var withSession = (eventStore, callback) => {
  const sessionFactory = canCreateEventStoreSession(eventStore) ? eventStore : nulloSessionFactory(eventStore);
  return sessionFactory.withSession(callback);
};

// src/commandHandling/handleCommandWithDecider.ts
var DeciderCommandHandler = (options) => async (eventStore, id, commands, handleOptions) => {
  const { decide, ...rest } = options;
  const deciders = (Array.isArray(commands) ? commands : [commands]).map(
    (command2) => (state) => decide(command2, state)
  );
  return CommandHandler(rest)(
    eventStore,
    id,
    deciders,
    handleOptions
  );
};

// src/messageBus/index.ts
var getInMemoryMessageBus = () => {
  const allHandlers = /* @__PURE__ */ new Map();
  let pendingMessages = [];
  return {
    send: async (command2) => {
      const handlers = allHandlers.get(command2.type);
      if (handlers === void 0 || handlers.length === 0)
        throw new EmmettError(
          `No handler registered for command ${command2.type}!`
        );
      const commandHandler = handlers[0];
      await commandHandler(command2);
    },
    publish: async (event2) => {
      const handlers = allHandlers.get(event2.type) ?? [];
      for (const handler of handlers) {
        const eventHandler = handler;
        await eventHandler(event2);
      }
    },
    schedule: (message2, when) => {
      pendingMessages = [...pendingMessages, { message: message2, options: when }];
    },
    handle: (commandHandler, ...commandTypes) => {
      const alreadyRegistered = [...allHandlers.keys()].filter(
        (registered) => commandTypes.includes(registered)
      );
      if (alreadyRegistered.length > 0)
        throw new EmmettError(
          `Cannot register handler for commands ${alreadyRegistered.join(", ")} as they're already registered!`
        );
      for (const commandType of commandTypes) {
        allHandlers.set(commandType, [
          commandHandler
        ]);
      }
    },
    subscribe(eventHandler, ...eventTypes) {
      for (const eventType of eventTypes) {
        if (!allHandlers.has(eventType)) allHandlers.set(eventType, []);
        allHandlers.set(eventType, [
          ...allHandlers.get(eventType) ?? [],
          eventHandler
        ]);
      }
    },
    dequeue: () => {
      const pending = pendingMessages;
      pendingMessages = [];
      return pending;
    }
  };
};

// src/processors/processors.ts
var getCheckpoint = (message2) => {
  return "checkpoint" in message2.metadata && // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
  isBigint(message2.metadata.checkpoint) ? (
    // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
    message2.metadata.checkpoint
  ) : "globalPosition" in message2.metadata && // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
  isBigint(message2.metadata.globalPosition) ? (
    // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
    message2.metadata.globalPosition
  ) : "streamPosition" in message2.metadata && // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
  isBigint(message2.metadata.streamPosition) ? (
    // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
    message2.metadata.streamPosition
  ) : null;
};
var wasMessageHandled = (message2, checkpoint) => {
  const messageCheckpoint = getCheckpoint(message2);
  const checkpointBigint = checkpoint;
  return messageCheckpoint !== null && messageCheckpoint !== void 0 && checkpointBigint !== null && checkpointBigint !== void 0 && messageCheckpoint <= checkpointBigint;
};
var MessageProcessorType = {
  PROJECTOR: "projector",
  REACTOR: "reactor"
};
var MessageProcessor = {
  result: {
    skip: (options) => ({
      type: "SKIP",
      ...options ?? {}
    }),
    stop: (options) => ({
      type: "STOP",
      ...options ?? {}
    })
  }
};
var defaultProcessingMessageProcessingScope = (handler, partialContext) => handler(partialContext);
var reactor = (options) => {
  const eachMessage = "eachMessage" in options && options.eachMessage ? options.eachMessage : () => Promise.resolve();
  let isActive = true;
  const { checkpoints, processorId, partition } = options;
  const processingScope = options.processingScope ?? defaultProcessingMessageProcessingScope;
  let lastCheckpoint = null;
  return {
    id: options.processorId,
    type: options.type ?? MessageProcessorType.REACTOR,
    close: () => options.hooks?.onClose ? options.hooks?.onClose() : Promise.resolve(),
    start: async (startOptions) => {
      isActive = true;
      if (lastCheckpoint !== null)
        return {
          lastCheckpoint
        };
      return await processingScope(async (context) => {
        if (options.hooks?.onStart) {
          await options.hooks?.onStart(context);
        }
        if (options.startFrom !== "CURRENT" && options.startFrom)
          return options.startFrom;
        if (checkpoints) {
          const readResult = await checkpoints?.read(
            {
              processorId,
              partition
            },
            { ...startOptions, ...context }
          );
          lastCheckpoint = readResult.lastCheckpoint;
        }
        if (lastCheckpoint === null) return "BEGINNING";
        return {
          lastCheckpoint
        };
      }, startOptions);
    },
    get isActive() {
      return isActive;
    },
    handle: async (messages, partialContext) => {
      if (!isActive) return Promise.resolve();
      return await processingScope(async (context) => {
        let result = void 0;
        for (const message2 of messages) {
          if (wasMessageHandled(message2, lastCheckpoint)) continue;
          const messageProcessingResult = await eachMessage(message2, context);
          if (checkpoints) {
            const storeCheckpointResult = await checkpoints.store(
              {
                processorId: options.processorId,
                version: options.version,
                message: message2,
                lastCheckpoint,
                partition: options.partition
              },
              context
            );
            if (storeCheckpointResult.success) {
              lastCheckpoint = storeCheckpointResult.newCheckpoint;
            }
          }
          if (messageProcessingResult && messageProcessingResult.type === "STOP") {
            isActive = false;
            result = messageProcessingResult;
            break;
          }
          if (options.stopAfter && options.stopAfter(message2)) {
            isActive = false;
            result = { type: "STOP", reason: "Stop condition reached" };
            break;
          }
          if (messageProcessingResult && messageProcessingResult.type === "SKIP")
            continue;
        }
        return result;
      }, partialContext);
    }
  };
};
var projector = (options) => {
  const { projection: projection2, ...rest } = options;
  return reactor({
    ...rest,
    type: MessageProcessorType.PROJECTOR,
    processorId: options.processorId ?? `projection:${projection2.name}`,
    hooks: {
      onStart: options.truncateOnStart && options.projection.truncate || options.hooks?.onStart ? async (context) => {
        if (options.truncateOnStart && options.projection.truncate)
          await options.projection.truncate(context);
        if (options.hooks?.onStart) await options.hooks?.onStart(context);
      } : void 0,
      onClose: options.hooks?.onClose
    },
    eachMessage: async (event2, context) => {
      if (!projection2.canHandle.includes(event2.type)) return;
      await projection2.handle([event2], context);
    }
  });
};

// src/processors/inMemoryProcessors.ts
var inMemoryCheckpointer = () => {
  return {
    read: async ({ processorId }, { database }) => {
      const checkpoint = await database.collection("emt_processor_checkpoints").findOne((d) => d._id === processorId);
      return Promise.resolve({
        lastCheckpoint: checkpoint?.lastCheckpoint ?? null
      });
    },
    store: async (context, { database }) => {
      const { message: message2, processorId, lastCheckpoint } = context;
      const checkpoints = database.collection(
        "emt_processor_checkpoints"
      );
      const checkpoint = await checkpoints.findOne(
        (d) => d._id === processorId
      );
      const currentPosition = checkpoint?.lastCheckpoint ?? null;
      const newCheckpoint = getCheckpoint(message2);
      if (currentPosition && (currentPosition === newCheckpoint || currentPosition !== lastCheckpoint)) {
        return {
          success: false,
          reason: currentPosition === newCheckpoint ? "IGNORED" : "MISMATCH"
        };
      }
      await checkpoints.handle(processorId, (existing) => ({
        ...existing ?? {},
        _id: processorId,
        lastCheckpoint: newCheckpoint
      }));
      return { success: true, newCheckpoint };
    }
  };
};
var inMemoryProcessingScope = (options) => {
  const processorDatabase = options.database;
  const processingScope = (handler, partialContext) => {
    const database = processorDatabase ?? partialContext?.database;
    if (!database)
      throw new EmmettError(
        `InMemory processor '${options.processorId}' is missing database. Ensure that you passed it through options`
      );
    return handler({ ...partialContext, database });
  };
  return processingScope;
};
var inMemoryProjector = (options) => {
  const database = options.connectionOptions?.database ?? getInMemoryDatabase();
  const hooks = {
    onStart: options.hooks?.onStart,
    onClose: options.hooks?.onClose ? async () => {
      if (options.hooks?.onClose) await options.hooks?.onClose();
    } : void 0
  };
  return {
    ...projector({
      ...options,
      hooks,
      processingScope: inMemoryProcessingScope({
        database,
        processorId: options.processorId ?? `projection:${options.projection.name}`
      }),
      checkpoints: inMemoryCheckpointer()
    }),
    database
  };
};
var inMemoryReactor = (options) => {
  const database = options.connectionOptions?.database ?? getInMemoryDatabase();
  const hooks = {
    onStart: options.hooks?.onStart,
    onClose: options.hooks?.onClose ? async () => {
      if (options.hooks?.onClose) await options.hooks?.onClose();
    } : void 0
  };
  return {
    ...reactor({
      ...options,
      hooks,
      processingScope: inMemoryProcessingScope({
        database,
        processorId: options.processorId
      }),
      checkpoints: inMemoryCheckpointer()
    }),
    database
  };
};

// src/projections/index.ts
var filterProjections = (type, projections2) => {
  const inlineProjections2 = projections2.filter((projection2) => projection2.type === type).map(({ projection: projection2 }) => projection2);
  const duplicateRegistrations = arrayUtils.getDuplicates(
    inlineProjections2,
    (proj) => proj.name
  );
  if (duplicateRegistrations.length > 0) {
    throw new EmmettError(`You cannot register multiple projections with the same name (or without the name).
      Ensure that:
      ${JSONParser.stringify(duplicateRegistrations)}
      have different names`);
  }
  return inlineProjections2;
};
var projection = (definition) => definition;
var inlineProjections = (definitions) => definitions.map((definition) => ({
  type: "inline",
  projection: definition
}));
var asyncProjections = (definitions) => definitions.map((definition) => ({
  type: "inline",
  projection: definition
}));
var projections = {
  inline: inlineProjections,
  async: asyncProjections
};

// src/streaming/binaryArrays.ts
var concatUint8Arrays = (chunks) => {
  const totalLength = chunks.reduce((acc, chunk) => acc + chunk.length, 0);
  const result = new Uint8Array(totalLength);
  let offset = 0;
  for (const chunk of chunks) {
    result.set(chunk, offset);
    offset += chunk.length;
  }
  return result;
};

// src/streaming/collectors/collect.ts
var collect = async (stream) => {
  const results = [];
  for await (const value of stream) {
    results.push(value);
  }
  return results;
};

// src/streaming/decoders/binary.ts
var BinaryJsonDecoder = class {
  buffer = [];
  addToBuffer(data) {
    this.buffer.push(data);
  }
  clearBuffer() {
    this.buffer = [];
  }
  hasCompleteMessage() {
    const combined = concatUint8Arrays(this.buffer);
    const text = new TextDecoder().decode(combined);
    return text.includes("\n");
  }
  decode() {
    if (!this.hasCompleteMessage()) {
      return null;
    }
    const combined = concatUint8Arrays(this.buffer);
    const text = new TextDecoder().decode(combined);
    const delimiterIndex = text.indexOf("\n");
    if (delimiterIndex === -1) {
      return null;
    }
    const jsonString = text.slice(0, delimiterIndex);
    const remaining = new Uint8Array(combined.buffer, delimiterIndex + 1);
    this.buffer = remaining.byteLength > 0 ? [remaining] : [];
    return JSON.parse(jsonString);
  }
};

// src/streaming/decoders/string.ts
var StringDecoder = class {
  constructor(transform) {
    this.transform = transform;
    this.transform = transform;
  }
  buffer = [];
  addToBuffer(data) {
    this.buffer.push(data);
  }
  clearBuffer() {
    this.buffer = [];
  }
  hasCompleteMessage() {
    return this.buffer.some((chunk) => chunk.includes("\n"));
  }
  decode() {
    const completeString = this.buffer.join("");
    if (!this.hasCompleteMessage()) {
      if (completeString.trim().length > 0) {
        throw new Error("Unterminated string in JSON at position");
      }
      return null;
    }
    const delimiterIndex = completeString.indexOf("\n");
    const message2 = completeString.slice(0, delimiterIndex).trim();
    this.buffer = [completeString.slice(delimiterIndex + 1)];
    return this.transform(message2);
  }
};

// src/streaming/decoders/json.ts
var JsonDecoder = class extends StringDecoder {
  constructor() {
    super((jsonString) => JSON.parse(jsonString));
  }
};

// src/streaming/decoders/object.ts
var ObjectDecoder = class {
  buffer = null;
  addToBuffer(data) {
    this.buffer = data;
  }
  clearBuffer() {
    this.buffer = null;
  }
  hasCompleteMessage() {
    return this.buffer !== null;
  }
  decode() {
    if (!this.hasCompleteMessage() || !this.buffer) {
      return null;
    }
    const data = this.buffer;
    this.clearBuffer();
    return data;
  }
};

// src/streaming/decoders/composite.ts
var CompositeDecoder = class {
  constructor(decoders) {
    this.decoders = decoders;
  }
  decoderFor(data) {
    const decoder = this.decoders.find((d) => d[0](data));
    if (!decoder) return null;
    return decoder[1];
  }
  addToBuffer(data) {
    this.decoderFor(data)?.addToBuffer(data);
  }
  clearBuffer() {
    for (const decoder of this.decoders.map((d) => d[1])) {
      decoder.clearBuffer();
    }
  }
  hasCompleteMessage() {
    return this.decoders.some((d) => d[1].hasCompleteMessage());
  }
  decode() {
    const decoder = this.decoders.map((d) => d[1]).find((d) => d.hasCompleteMessage());
    return decoder?.decode() ?? null;
  }
};
var DefaultDecoder = class extends CompositeDecoder {
  constructor() {
    super([
      [(data) => typeof data === "string", new JsonDecoder()],
      [(data) => data instanceof Uint8Array, new BinaryJsonDecoder()],
      [(data) => typeof data === "object", new ObjectDecoder()]
    ]);
  }
};

// src/streaming/generators/fromArray.ts
import { ReadableStream } from "web-streams-polyfill";
var fromArray = (chunks) => new ReadableStream({
  start(controller) {
    for (const chunk of chunks) controller.enqueue(chunk);
    controller.close();
  }
});

// src/streaming/generators/index.ts
var streamGenerators = { fromArray };

// src/streaming/restream.ts
import "web-streams-polyfill";

// src/streaming/transformations/filter.ts
import { TransformStream as TransformStream3 } from "web-streams-polyfill";
var filter = (filter2) => new TransformStream3({
  transform(chunk, controller) {
    if (filter2(chunk)) {
      controller.enqueue(chunk);
    }
  }
});

// src/streaming/transformations/map.ts
import { TransformStream as TransformStream4 } from "web-streams-polyfill";
var map = (map2) => new TransformStream4({
  transform(chunk, controller) {
    controller.enqueue(map2(chunk));
  }
});

// src/streaming/transformations/reduce.ts
import { TransformStream as TransformStream5 } from "web-streams-polyfill";
var reduce = (reducer, initialValue) => new ReduceTransformStream(reducer, initialValue);
var ReduceTransformStream = class extends TransformStream5 {
  accumulator;
  reducer;
  constructor(reducer, initialValue) {
    super({
      transform: (chunk) => {
        this.accumulator = this.reducer(this.accumulator, chunk);
      },
      flush: (controller) => {
        controller.enqueue(this.accumulator);
        controller.terminate();
      }
    });
    this.accumulator = initialValue;
    this.reducer = reducer;
  }
};

// src/streaming/transformations/retry.ts
import {
  TransformStream as TransformStream6
} from "web-streams-polyfill";
var retryStream = (createSourceStream, handleChunk2, retryOptions = { forever: true, minTimeout: 25 }) => new TransformStream6({
  start(controller) {
    asyncRetry(
      () => onRestream(createSourceStream, handleChunk2, controller),
      retryOptions
    ).catch((error2) => {
      controller.error(error2);
    });
  }
});
var onRestream = async (createSourceStream, handleChunk2, controller) => {
  const sourceStream = createSourceStream();
  const reader = sourceStream.getReader();
  try {
    let done;
    do {
      const result = await reader.read();
      done = result.done;
      await handleChunk2(result, controller);
      if (done) {
        controller.terminate();
      }
    } while (!done);
  } finally {
    reader.releaseLock();
  }
};

// src/streaming/transformations/skip.ts
import { TransformStream as TransformStream7 } from "web-streams-polyfill";
var skip = (limit) => new SkipTransformStream(limit);
var SkipTransformStream = class extends TransformStream7 {
  count = 0;
  skip;
  constructor(skip2) {
    super({
      transform: (chunk, controller) => {
        this.count++;
        if (this.count > this.skip) {
          controller.enqueue(chunk);
        }
      }
    });
    this.skip = skip2;
  }
};

// src/streaming/transformations/stopAfter.ts
import { TransformStream as TransformStream8 } from "web-streams-polyfill";
var stopAfter = (stopCondition) => new TransformStream8({
  transform(chunk, controller) {
    controller.enqueue(chunk);
    if (stopCondition(chunk)) {
      controller.terminate();
    }
  }
});

// src/streaming/transformations/stopOn.ts
import { TransformStream as TransformStream9 } from "web-streams-polyfill";
var stopOn = (stopCondition) => new TransformStream9({
  async transform(chunk, controller) {
    if (!stopCondition(chunk)) {
      controller.enqueue(chunk);
      return;
    }
    await Promise.resolve();
    controller.terminate();
  }
});

// src/streaming/transformations/take.ts
import { TransformStream as TransformStream10 } from "web-streams-polyfill";
var take = (limit) => new TakeTransformStream(limit);
var TakeTransformStream = class extends TransformStream10 {
  count = 0;
  limit;
  constructor(limit) {
    super({
      transform: (chunk, controller) => {
        if (this.count < this.limit) {
          this.count++;
          controller.enqueue(chunk);
        } else {
          controller.terminate();
        }
      }
    });
    this.limit = limit;
  }
};

// src/streaming/transformations/waitAtMost.ts
import { TransformStream as TransformStream11 } from "web-streams-polyfill";
var waitAtMost = (waitTimeInMs) => new TransformStream11({
  start(controller) {
    const timeoutId = setTimeout(() => {
      controller.terminate();
    }, waitTimeInMs);
    const originalTerminate = controller.terminate.bind(controller);
    controller.terminate = () => {
      clearTimeout(timeoutId);
      originalTerminate();
    };
  },
  transform(chunk, controller) {
    controller.enqueue(chunk);
  }
});

// src/streaming/transformations/index.ts
var streamTransformations = {
  filter,
  take,
  TakeTransformStream,
  skip,
  SkipTransformStream,
  map,
  notifyAboutNoActiveReadersStream,
  NotifyAboutNoActiveReadersStream,
  reduce,
  ReduceTransformStream,
  retry: retryStream,
  stopAfter,
  stopOn,
  waitAtMost
};

// src/streaming/restream.ts
var { retry: retry2 } = streamTransformations;
var restream = (createSourceStream, transform = (source) => source, retryOptions = { forever: true, minTimeout: 25 }, decoder = new DefaultDecoder()) => retry2(createSourceStream, handleChunk(transform, decoder), retryOptions).readable;
var handleChunk = (transform = (source) => source, decoder = new DefaultDecoder()) => (readResult, controller) => {
  const { done: isDone, value } = readResult;
  if (value) decoder.addToBuffer(value);
  if (!isDone && !decoder.hasCompleteMessage()) return;
  decodeAndTransform(decoder, transform, controller);
};
var decodeAndTransform = (decoder, transform, controller) => {
  try {
    const decoded = decoder.decode();
    if (!decoded) return;
    const transformed = transform(decoded);
    controller.enqueue(transformed);
  } catch (error2) {
    controller.error(new Error(`Decoding error: ${error2?.toString()}`));
  }
};
export {
  AssertionError,
  BinaryJsonDecoder,
  CaughtUpTransformStream,
  CommandHandler,
  CommandHandlerStreamVersionConflictRetryOptions,
  CompositeDecoder,
  ConcurrencyError,
  ConcurrencyInMemoryDatabaseError,
  DATABASE_REQUIRED_ERROR_MESSAGE,
  DeciderCommandHandler,
  DeciderSpecification,
  DefaultDecoder,
  EmmettError,
  ExpectedVersionConflictError,
  GlobalStreamCaughtUpType,
  IllegalStateError,
  InMemoryEventStoreDefaultStreamVersion,
  InMemoryProjectionSpec,
  InProcessLock,
  JSONParser,
  JsonDecoder,
  MessageProcessor,
  MessageProcessorType,
  NO_CONCURRENCY_CHECK,
  NoRetries,
  NotFoundError,
  ObjectDecoder,
  ParseError,
  STREAM_DOES_NOT_EXIST,
  STREAM_EXISTS,
  StreamingCoordinator,
  StringDecoder,
  TaskProcessor,
  ValidationError,
  ValidationErrors,
  WrapEventStore,
  accept,
  argMatches,
  argValue,
  arrayUtils,
  assertDeepEqual,
  assertDoesNotThrow,
  assertEqual,
  assertExpectedVersionMatchesCurrent,
  assertFails,
  assertFalse,
  assertIsNotNull,
  assertIsNull,
  assertMatches,
  assertNotDeepEqual,
  assertNotEmptyString,
  assertNotEqual,
  assertOk,
  assertPositiveNumber,
  assertRejects,
  assertThat,
  assertThatArray,
  assertThrows,
  assertThrowsAsync,
  assertTrue,
  assertUnsignedBigInt,
  asyncAwaiter,
  asyncProjections,
  asyncRetry,
  canCreateEventStoreSession,
  caughtUpEventFrom,
  collect,
  command,
  complete,
  concatUint8Arrays,
  deepEquals,
  defaultProcessingMessageProcessingScope,
  delay,
  documentExists,
  error,
  event,
  eventInStream,
  eventsInStream,
  expectInMemoryDocuments,
  filterProjections,
  formatDateToUtcYYYYMMDD,
  forwardToMessageBus,
  getCheckpoint,
  getInMemoryDatabase,
  getInMemoryEventStore,
  getInMemoryMessageBus,
  globalStreamCaughtUp,
  handleInMemoryProjections,
  ignore,
  inMemoryCheckpointer,
  inMemoryMultiStreamProjection,
  inMemoryProjection,
  inMemoryProjector,
  inMemoryReactor,
  inMemorySingleStreamProjection,
  inlineProjections,
  isBigint,
  isEquatable,
  isErrorConstructor,
  isExpectedVersionConflictError,
  isGlobalStreamCaughtUp,
  isNotInternalEvent,
  isNumber,
  isPluginConfig,
  isString,
  isSubscriptionEvent,
  isSubset,
  isValidYYYYMMDD,
  matchesExpectedVersion,
  merge,
  message,
  newEventsInStream,
  nulloSessionFactory,
  parseDateFromUtcYYYYMMDD,
  projection,
  projections,
  projector,
  publish,
  reactor,
  reply,
  restream,
  schedule,
  send,
  streamGenerators,
  streamTrackingGlobalPosition,
  streamTransformations,
  sum,
  tryPublishMessagesAfterCommit,
  verifyThat,
  wasMessageHandled
};
//# sourceMappingURL=index.js.map